import time
import numpy as np
import numba
from numba import jit,njit, prange, types
import warnings
from kernelml import KernelML
import sys
import copy
from functools import partial
from .hdr_helpers_bycython import *

class ArgumentError(Exception):
    pass

class VersionError(Exception):
    pass

    
warnings.filterwarnings('ignore')    

### HDRE
class DensityFactorization():

    def __init__(self,number_of_clusters,bins_per_dimension=41,smoothing_parameter=3.0):
        self.num_clusters = number_of_clusters
        self.smoothing_parameter = smoothing_parameter
        self.bins_per_dim = bins_per_dimension
        self.kmldata = None
        self.kde_target = None
        self.norm = 1

        half = (self.bins_per_dim)//2
        mesh = np.meshgrid(*[np.arange(-(half),half+1,1) for _ in range(2)])
        mesh = [d**2 for d in mesh]
        sigma = self.smoothing_parameter
        kernel = np.exp(-sum(mesh)/(2*sigma**2))/np.sqrt(2*np.pi*sigma**2)
        kernel = kernel/np.sum(kernel)
        self.fftkernel = np.fft.fftn(kernel)
        
        
    def load_kde(self,kde):
        self.kde_target = kde
        
    def update_kde(self,X,y=None,alpha=0.5):
        if self.kde_target is not None:
        
            combo_len = len(self.dim_combos)
            
            count=0
            _y_ = None
            for i,j in self.dim_combos:
                _X_ = X[:,[i,j]]
                bins3 = [self.dim_bins[i],self.dim_bins[j]]
                if y is not None and self.multi_dim==True:
                    _y_ = y[:,[i,j]]
                data,_ = hdre_histogram(_X_,y,agg_func='count',cost=None,bins=bins3)
                data=data/np.sum(data)
                new_kde = np.fft.fftshift(np.real(np.fft.ifftn(np.fft.fftn(data)*self.fftkernel)))
                self.kde_target[count]=self.kde_target[count]*(1-alpha) + new_kde*(alpha)
                count+=1

    def optimize(self,X,y=None,cost=None, number_of_random_simulations=1000, number_of_realizations=10, agg_func='count',verbose=True,
        multi_dim=False,
        ):
        
        self.simulations = number_of_random_simulations
        self.realizations = number_of_realizations
        
        if (self.bins_per_dim/2 == self.bins_per_dim//2+1):
            raise ArgumentError("The number of bins per dimensions must be an odd integer")
        
        self.num_dim = X.shape[1]
        self.multi_dim=multi_dim
        
        maxs = np.max(X,axis=0)
        mins = np.min(X,axis=0)
        widths = (maxs-mins)/(self.bins_per_dim)
        max_lim = maxs.copy()
        min_lim = mins.copy()
        maxs+=int(np.sqrt(self.bins_per_dim))*widths
        mins-=int(np.sqrt(self.bins_per_dim))*widths
        widths = (maxs-mins)/(self.bins_per_dim)
        
        self.dim_bins = [np.linspace(m0,m1,self.bins_per_dim) for m0,m1 in zip(mins,maxs)]
        
        self.dim_combos = [(i,j) for i in range(X.shape[1]) for j in range(X.shape[1]) if j>i]
        #custom dimensional pair TODO
        #if dim_tuples is None:
        #    self.dim_combos = [(i,j) for i in range(X.shape[1]) for j in range(X.shape[1]) if j>i]
        #else:
        #    self.dim_combos=dim_tuples
        
        combo_len = len(self.dim_combos)
        
        _y_ = None
        if self.kde_target is None:
            self.kde_target = np.zeros((combo_len,self.bins_per_dim,self.bins_per_dim),dtype=np.float64)
            self.bin_combos = np.zeros((combo_len,self.bins_per_dim,2),dtype=np.float64)
            count=0
            for i,j in self.dim_combos:
                _X_ = X[:,[i,j]]
                if y is not None and self.multi_dim==True:
                    _y_ = y[:,[i,j]]

                bins3 = [self.dim_bins[i],self.dim_bins[j]]
    #             bins3 = [np.concatenate([[-np.inf],_bins_,[np.inf]]) for _bins_ in bins3]
                self.bin_combos[count] = np.column_stack(bins3)
                data,_ = hdre_histogram(_X_,y,agg_func=agg_func,cost=cost,bins=bins3)
                data=data/np.sum(data)
                self.kde_target[count] = np.fft.fftshift(np.real(np.fft.ifftn(np.fft.fftn(data)*self.fftkernel)))

                count+=1

        
        cycles = 100

        #The number of total simulations per realization = number of cycles * numer of simulations

        zcore = 2.0
        volume = 10 + self.num_dim
        volatility = 1
        zscore = 1


        param_to_dim = np.arange(0,self.num_dim*self.num_clusters)%self.num_dim
        param_to_dim = param_to_dim.astype(np.float)
        self.dim_combos = np.array(self.dim_combos).astype(np.int64)
        args = [self.dim_combos,self.kde_target,self.bin_combos,
                    min_lim,max_lim,param_to_dim,self.num_dim,widths]
        self.args = args
        
        if self.kmldata is None:
            self.kml = KernelML(
                 prior_sampler_fcn=hdre_prior_sampler,
                 posterior_sampler_fcn=None,
                 intermediate_sampler_fcn=None,
                 mini_batch_sampler_fcn=None,
                 parameter_transform_fcn=hdre_parameter_transform,
                 loss_calculation_fcn=_map_losses,
                 batch_size=None)
        else:
            self.kmldata.args = self.args
            self.kml.load_kmldata(self.kmldata)
#            self.kml.kmldata.args = self.args
            
        self.kml.optimize(X[:1],np.array([[]]),
                                        convergence_z_score=3.0,
                                        min_loss_per_change=0.0,
                                        number_of_parameters=self.num_clusters*self.num_dim+self.num_dim,
                                        args=args,
                                        number_of_realizations=self.realizations,
                                        number_of_random_simulations=self.simulations,
                                        update_volume=volume,
                                        update_volatility=volatility,
                                        number_of_cycles=cycles,
                                        print_feedback=verbose)

        self.kmldata = self.kml.kmldata
        self.kml.load_kmldata(self.kmldata)
        
        self.kde_estimate = np.zeros((combo_len,self.bins_per_dim,self.bins_per_dim))
        var = self.deviations_
        mean = self.centroids_

        count=0
        for i,j in self.dim_combos:
            bins = self.bin_combos[count:count+1,:,0].shape[1]
            bins3 = [self.bin_combos[count:count+1,:,0],self.bin_combos[count:count+1,:,1]]
            data1 = np.random.uniform(0,1e-3,size=(bins,bins))
            for k in range(self.num_clusters):

                pdf = uniform_kernel(bins3[0],mean[k,i],var[i,0]).dot(uniform_kernel(bins3[1],mean[k,j],var[j,0]).T)
                data1 += pdf

            data1=data1/np.sum(data1)
#             pdf = np.fft.fftshift(np.real(np.fft.ifftn(np.fft.fftn(data1)*fftkernel)))
            self.kde_estimate[count] = data1
            
#             self.kde_target[count] = np.flipud(self.kde_target[count])
            count+=1
    
    @property
    def deviations_(self):
        w = self.kmldata.best_weight_vector.flatten()
        return np.abs(w[:self.num_dim]).reshape(-1,1)

    @property
    def centroids_(self):
        w = self.kmldata.best_weight_vector.flatten()
        w = w[self.num_dim:]
        return w.reshape((w.size//self.num_dim,self.num_dim))
    
    def prune_clusters(self,X,pad=1.0,limit=0):
        mask=self.get_assignments(X,pad=pad)
        reduce = [i for i in range(mask.shape[1]) if np.sum(mask[:,i]==1,axis=0)<=limit]
        
        
        mean = np.delete(self.centroids_, reduce, axis=0).flatten()
        variance = self.deviations_.flatten()
        
        self.kmldata.best_weight_vector = np.concatenate([variance,mean]).reshape(-1,1)
        
       
        
        centroid_indxs = np.arange(0,self.num_dim*self.num_clusters)
        param_to_clusters = centroid_indxs%self.num_clusters
        self.num_clusters = self.num_clusters - len(reduce)
        self.kmldata.num_clusters = self.num_clusters
    
        reduce = [x+self.num_dim for x in centroid_indxs if param_to_clusters[x] in reduce]
        
        self.kmldata.update_by_cycle = np.delete(self.kmldata.update_by_cycle, reduce, axis=1)
        
        self.kmldata.update_by_realization = np.delete(self.kmldata.update_by_realization, reduce, axis=1)
    
        self.kmldata.update_history = np.delete(self.kmldata.update_history, reduce, axis=0)
        
        self.kmldata.current_weights = np.delete(self.kmldata.current_weights, reduce, axis=0)
        
        self.kmldata.number_of_parameters =self.num_clusters*self.num_dim+self.num_dim
        
        self.get_kde_estimate(pad)
        
    def get_kde_estimate(self,pad=1.0):
        
        dim_combos,pdf_combos,bin_combos,min_lim,max_lim,param_to_dim,num_dim,widths = self.kmldata.args
        combo_len = len(self.dim_combos)

        self.kde_estimate = np.zeros((combo_len,self.bins_per_dim,self.bins_per_dim))
        var = self.deviations_
        mean = self.centroids_
        count=0
        for i,j in self.dim_combos:
            bins = bin_combos[count:count+1,:,0].shape[1]
            bins3 = [bin_combos[count:count+1,:,0],bin_combos[count:count+1,:,1]]
            
            data1 = np.random.uniform(0,1e-3,size=(bins,bins))
            for k in range(self.num_clusters):

                pdf = uniform_kernel(bins3[0],mean[k,i],var[i,0]).dot(uniform_kernel(bins3[1],mean[k,j],var[j,0]).T)
                data1 += pdf

            data1=data1/np.sum(data1)
            self.kde_estimate[count] = data1
            count+=1
    

    def get_polygons(self,i,j,k,pad=1):
        var = self.deviations_
        mean = self.centroids_

        S = np.array([[var[i],0],[0,var[j]]])

        T = (S*pad)

        points = np.array([[-1.,1.],[1.,1.],[1.,-1.],[-1.,-1.],[-1.,1.]])
        points = points.dot(T)
        points[:,0] = points[:,0]+mean[k,i]
        points[:,1] = points[:,1]+mean[k,j]


        return points


    def get_assignments(self,X,pad=1):
        var = self.deviations_
        mean = self.centroids_
        mask = np.zeros((X.shape[0],self.num_clusters),dtype=np.bool)
        count=0
        for k in range(self.num_clusters):
            lower = mean[k].flatten()-var.flatten()*pad
            upper = mean[k].flatten()+var.flatten()*pad

            mask[:,k] = np.all((X>=lower)&(X<=upper),axis=1)

        return mask
    

    def get_distances(self,X,pad=1,distance='chebyshev'):
        var = self.deviations_
        w = self.centroids_

        loss_matrix=np.zeros((X.shape[0],self.num_clusters))
        for i in range(w.shape[0]):
            if distance=='chebyshev':
                loss_matrix[:,i]=np.max(np.abs(X[:]-w[i]),axis=1)
            elif distance=='euclidean':
                loss_matrix[:,i]=np.sum(np.abs(X[:]-w[i])**2,axis=1)
            elif distance=='mae':
                loss_matrix[:,i]=np.mean(np.abs(X[:]-w[i]),axis=1)
            else:
                raise ArgumentError("Invalid distance metric")
        return loss_matrix


class HierarchicalDensityFactorization():
    
    def __init__(self,num_clusters=8,
                 bins_per_dimension=61,
                 smoothing_parameter=1.0,
                 min_leaf_samples=1,
                 agg_func='count',
                 multi_dim=False,
                 alpha = 0.1,
                 verbose=0):
        
        self.alpha = alpha
        self.min_leaf_samples = min_leaf_samples
        self.agg_func = agg_func
        self.multi_dim = multi_dim
        self.assignments = []
        self.centroids = []
        self.deviations = []
        self.min_losses = []
        self.models = []
        self.burn_in = 10
        self.verbose = verbose
        self.kde_true = []
        self.kde_pred = []
        self.KDE = None
        
        self.smoothing_parameter = smoothing_parameter
        
        self.model = DensityFactorization(number_of_clusters=num_clusters,
                                                                    bins_per_dimension=bins_per_dimension,
                                                                    smoothing_parameter=smoothing_parameter)
        self.count = 0
        
        
 
    def refactor_assignments(self,X,pad=1.):
        
        assignments = []
        for model in self.models:
            
            D = model.get_assignments(X,pad=1.).astype(np.int)
            
            Z = np.diag(np.dot(D.T,D))
            for i in range(Z.shape[0]):
                if Z[i]>self.min_leaf_samples:
                    assignments.append(D[:,i:i+1])
                    
#             E = np.hstack(self.assignments)
#             M = (np.sum(E,axis=1)==0)

#             alpha = np.mean(M)*self.alpha

#             self.model.update_kde(X[M],X[M],alpha)
            
        return assignments

#    def assign(self,X,pad=1.):
#        assignments = []
#        for model in self.models:
#
#            D = model.get_assignments(X,pad=1.).astype(np.int)
#
#            Z = np.diag(np.dot(D.T,D))
#            for i in range(Z.shape[0]):
#                if Z[i]>self.min_leaf_samples:
#                    assignments.append(D[:,i:i+1])
#
#        return np.hstack(assignments)
       
    def assign(self,X,pad=1):
        num_clusters = np.vstack(self.centroids).shape[0]
        mask = np.zeros((X.shape[0],num_clusters),dtype=np.bool)
        k=0
        for C,S in zip(self.centroids,self.deviations):
            for i in range(C.shape[0]):
                lower = C[i].flatten()-S.flatten()*pad
                upper = C[i].flatten()+S.flatten()*pad
                mask[:,k] = np.all((X>=lower)&(X<=upper),axis=1)
                k+=1
                
        return mask
    
    def factorize(self,X,M,Y=None):

        
        """
        alpha controls how much weight to give to data points not assign to a cluster
        """
        
        alpha = np.mean(M)*self.alpha

        self.model.update_kde(X[M],X[M],alpha)
        
        realizations = self.realizations
            
        self.model.kmldata=None
        
#         model.prune_clusters(y,pad=2.0,limit=1)
        y=X
        if Y is not None:
            y=Y
            
        
        self.model.optimize(X=X,y=y,  number_of_random_simulations=self.number_of_random_simulations,
                   number_of_realizations = realizations,
                   agg_func=self.agg_func,
                   verbose=self.verbose,
                   multi_dim=self.multi_dim
                     )
        
#         if self.count==0:
#             self.KDE = self.model.kde_target.copy()
        
        self.count+=1
        
        #this will plot the estimated kernel density estimated
        """
        count=0
        for i,j in self.model.dim_combos:

            plt.subplot(1,2,1)
            plt.imshow(self.model.kde_estimate[count])
            plt.title('Estimated KDE')

            plt.subplot(1,2,2)
            plt.imshow(self.model.kde_target[count])
            plt.title('Target KDE')
            plt.show()
            count+=1
        """
            
        self.kde_true.append(self.model.kde_target.copy())
        self.kde_pred.append(self.model.kde_estimate.copy())
        

        D = self.model.get_assignments(X,pad=1.).astype(np.int)
        C = self.model.centroids_
        S = self.model.deviations_

        minloss = self.model.kmldata.loss_by_realization[-1][0]
        self.min_losses.append(minloss)
        
        self.models.append(copy.copy(self.model))
        
        #add data points to the collected list
        Z = np.diag(np.dot(D.T,D))
        
        if self.verbose==True:
            print(np.sum(Z==0),'clusters were not assigned data points\n')
        
        assigns = []
        centers = []
        for i in range(Z.shape[0]):
            #only add if the cluster has new data points?
            if Z[i]>self.min_leaf_samples:
                assigns.append(D[:,i:i+1])
                centers.append(C[i:i+1])
                
                
        if len(centers)>0:
            centers = np.vstack(centers)
            self.assignments.append(assigns)
            self.centroids.append(centers)
            self.deviations.append(S)
            
    def optimize(self,X,Y=None,
                 maxiter=20,
                 stop_thrshld=0.01,
                 realizations=10,
                 number_of_random_simulations=200,
                 dim_tuples=None,
                 verbose=0):
        
        self.verbose = verbose
        self.dim_tuples = dim_tuples
        self.realizations = realizations
        self.number_of_random_simulations = number_of_random_simulations
        M = np.ones(X.shape[0],dtype=bool)
        LAST = M
        for _ in range(maxiter):
            self.factorize(X,M,Y)
            E = np.hstack(sum(self.assignments,[]))
            M = (np.sum(E,axis=1)==0)
            if self.verbose==True:
                print(np.sum(M),'data points are unassigned')
            
            #break when no data points are left
            if np.sum(M)==0:
                break
                
            #break if no threshold is met
                        
            
            prune_losses = []
            prune_kde_true = []
            prune_kde_pred = []
            prune_models = []
            prune_centroids = []
            prune_deviations = []
            prune_assignments = []
            keep_count=0
            
            if len(self.min_losses)<=1:
                continue
            
            
            for i,loss in enumerate(self.min_losses):
                if loss*0.5<=np.min(self.min_losses):
                    prune_models.append(self.models[i])
                    prune_kde_true.append(self.kde_true[i])
                    prune_kde_pred.append(self.kde_pred[i])
                    prune_losses.append(self.min_losses[i])
                    prune_centroids.append(self.centroids[i])
                    prune_deviations.append(self.deviations[i])
                    prune_assignments.append(self.assignments[i])
                    keep_count+=1
                    
            prune_count = len(self.min_losses)-keep_count

            if self.verbose==True and prune_count>0:
                print('pruning {} iterations\n'.format(prune_count))
                
            self.models = prune_models
            self.kde_true = prune_kde_true
            self.kde_pred = prune_kde_pred
            self.min_losses = prune_losses
            self.centroids = prune_centroids
            self.deviations = prune_deviations
            self.assignments = prune_assignments
            

#            if prune_count>0:
#                self.assignments = self.refactor_assignments(X)
            
            
            
            if np.mean(M)<stop_thrshld and prune_count==0:
                if self.verbose==True:
                    print('stopping with {} unassigned data points'.format(np.sum(M)))
                break
            
            if (np.sum(LAST)-np.sum(M))==0 and prune_count==0:
                if self.verbose==True:
                    print('stopping with {} unassigned data points'.format(np.sum(M)))
                break
                
            LAST = M
            
        if self.verbose==True and self.count==maxiter:
            print('stopping at {} iterations'.format(maxiter))
        

            
        
            

@jit('float64[:,:](float64[:,:], float64,float64)',nopython=True)
def uniform_kernel(x,mu_,var_):

    mu = mu_
    
    x = x.ravel()

    width = x[1]-x[0]
    var = var_
    
    if var<width:
        var = width
    
    pdf = ((x+width>=mu-var)&(x+width<=mu+var)).reshape(-1,1)
    pdf = pdf.astype(np.float64)
    
    if x[1]>mu-var:
        pdf[0] = pdf[0] + var*2/width - np.sum(pdf)
    if x[-2]<mu+var:
        pdf[-1] = pdf[-1] + var*2/width - np.sum(pdf)
    result = (pdf/np.sum(pdf))
    return result



@jit('float64(float64[:,:], float64[:,:], float64[:,:],int64[:,:],float64[:,:,:],float64[:,:,:],float64[:],float64[:],float64[:],int64,float64[:])',nopython=True)
def hdre_loss(X,y,w_,dim_combos, pdf_combos,bin_combos, min_lim, max_lim, param_to_dim, num_dim, widths):
     var =np.copy(np.abs(w_[:num_dim])).astype(np.float64)

     num_clusters = int(w_[num_dim:].size/num_dim)

     samples = 100
     w =  np.copy(w_[num_dim:]).reshape((num_clusters,num_dim)).astype(np.float64)

 #     assert num_clusters==w.size/dim

     N = dim_combos.shape[0]

     bin_combos2 = np.copy(bin_combos).astype(np.float64)
     ll=0
     
     count=0
     for n in range(N):
         i,j = dim_combos[n]
         binsx = bin_combos[count:count+1,:,0].shape[1]
         bins = bin_combos2[count:count+1,:,:].astype(np.float64)
         
         data1 = np.random.uniform(0,1e-3,size=(binsx,binsx))
         for k in range(w.shape[0]):
             bins0 = bins[:,:,0].astype(np.float64)
             bins1 = bins[:,:,1].astype(np.float64)
 #             uniform_kernel(bins0,w[k,i],var[i,0])
             pdf = uniform_kernel(bins0,w[k,i],var[i,0]).dot(uniform_kernel(bins1,w[k,j],var[j,0]).T)
          
 #             print(pdf.shape)
             data1 += pdf
             

         pdf1=data1/np.sum(data1)

         pdf2 = pdf_combos[count]

         diff = (pdf2-pdf1).flatten()*100
         
         ll += np.sum(diff**2)

         count+=1

     return ll

@njit('float64[:](float64[:,:], float64[:,:], float64[:,:],int64[:,:],float64[:,:,:],float64[:,:,:],float64[:],float64[:],float64[:],int64,float64[:])',parallel=True)
def _map_losses(X, y, w_list, dim_combos, pdf_combos,bin_combos, min_lim, max_lim, param_to_dim, num_dim, widths):
    N = w_list.shape[1]
    resX = np.zeros(N)
    for i in prange(N):
        loss = hdre_loss(X,y,w_list[:,i:i+1],dim_combos, pdf_combos,bin_combos, min_lim, max_lim, param_to_dim, num_dim, widths)
        resX[i] = loss
    return resX



import time
from sklearn import linear_model
import numpy as np
#import matplotlib.pyplot as plt
import heapq
import gc

try:
    import ipyparallel as ipp
except:
    pass

def get_parallel_losses(loss_function, X, y, weights, args, indexes):

    heap=[]
    for i in indexes:

        if len(args)>0:
            loss = loss_function(X,y,weights[:,i:i+1],list(args))
        else:
            loss = loss_function(X,y,weights[:,i:i+1])

        #make sure the loss is not infinity (this might be a problem for logistic regression)
        if abs(loss)<np.float('inf'):
            heap.append((loss,i))
    return heap

class CustomException(Exception):
    pass

class kernel_optimizer():

    def __init__(self,X,y,loss_function,number_of_parameters,args=[],shapes=[],model=None):
        self.X = X
        self.y = y
        self.loss_function = loss_function
        self.kmldata = model
        self.number_of_parameters = number_of_parameters
        self.args = np.array(args)
        self.shapes = np.array(shapes)
        self.loss_by_cycle = []
        self.update_by_cycle = []
        self.default_random_simulation_params()
        self.adjust_optimizer()
        self.prior_sampler = self.prior_sampler_uniform_distribution
        self.random_sampler = self.posterior_sampler_uniform_distribution
        self.parameter_transform = self.default_parameter_transform
        self.convergence_z_score = 2
        self.min_loss_per_change = 0
        self.use_ipyparallel = False
        self.num_cores = 1
        self.has_args = False


        if len(self.args)>0:
            self.has_args = True


    def set_weights(self,w):
        self.kmldata.current_weights = w

    def init_ipyparallel(self,dview,use_ipyparallel=True):
        self.num_cores = len(dview.targets)
        self.use_ipyparallel = use_ipyparallel
        self.dview = dview
        self.loss_function.__name__ = 'loss_function'
        self.dview['loss_function']=self.loss_function
#         self.dview['reshape_vector']=self.reshape_vector ,shapes=self.shapes.copy()
        mydict=dict(X=self.X.copy(),y=self.y.copy(),args=self.args.copy())
        self.dview.push(mydict)
        self.dview.execute('import gc')

    def heapsort(self,iterable):
        h = []
        for value in iterable:
            heapq.heappush(h, value)
        return [heapq.heappop(h) for i in range(len(h))]

    def adjust_convergence_z_score(self,z):
        self.convergence_z_score = z

    def adjust_min_loss_per_change(self,mlpc):
        self.min_loss_per_change = mlpc

    def default_parameter_transform(self,w,args):
        return w

    def default_random_simulation_params(self,prior_random_samples=None,random_sample_num=None,prior_uniform_low=-1,prior_uniform_high=1):
        self.kmldata.prior_random_samples = prior_random_samples
        self.kmldata.posterior_random_samples = random_sample_num
        self.low = prior_uniform_low
        self.high = prior_uniform_high

    def adjust_optimizer(self,number_of_random_simulations=100,update_volatility=1,number_of_cycles=10,
                                   update_volume=2,
                                   analyze_n_parameters=None
                                   ):

        self.number_of_random_simulations = number_of_random_simulations
        self.update_volatility = update_volatility
        self.number_of_cycles = number_of_cycles
        self.update_volume = update_volume
        self.strokes = update_volume

        self.analyze_n_parameters = analyze_n_parameters
        if self.analyze_n_parameters is None:
            self.analyze_n_parameters =  int(0.25*self.number_of_random_simulations)

        if self.kmldata.prior_random_samples is None:
            self.kmldata.prior_random_samples = int(self.number_of_random_simulations)

        if self.kmldata.posterior_random_samples is None:
            self.kmldata.posterior_random_samples = int(self.number_of_random_simulations)


        self.P = 0.9
        self.I = 0.9
        self.D = 0

    def _update_parameter_history(self):

        if len(self.kmldata.update_history) == 0:
            self.kmldata.update_history = self.kmldata.current_weights
            self.kmldata.loss_history = self.loss
            return
        try:
            self.kmldata.update_history = np.column_stack((self.kmldata.update_history,self.kmldata.current_weights))
        except Exception as e:
                raise CustomException("Not enough samples, try increasing the number of simulations or volatility")

        self.kmldata.loss_history =  np.hstack((self.kmldata.loss_history,self.loss))

        try:
            self.kmldata.update_history.shape[1]
            self.kmldata.update_history.shape[0]
        except Exception as e:
            raise CustomException('parameter input must be 2-D, please check sample size or number of parameters.')

    def _map_simulated_losses_parallel(self):

        weights_index = np.arange(0,self.kmldata.current_weights.shape[1])
        weights_index_split = np.array_split(weights_index,self.num_cores)

        mydict=dict(weights=self.kmldata.current_weights)
        self.dview.push(mydict)

        rX = [ipp.Reference('X') for i in range(self.num_cores)]
        ry = [ipp.Reference('y') for i in range(self.num_cores)]
        rweights = [ipp.Reference('weights') for i in range(self.num_cores)]
        rloss_function = [ipp.Reference('loss_function') for i in range(self.num_cores)]
        rargs = [ipp.Reference('args') for i in range(self.num_cores)]
        heap_array = self.dview.map_sync(get_parallel_losses,rloss_function,rX,ry,rweights,rargs,weights_index_split)
        heap =  []
        for val in heap_array:
            heap.extend(val)
        return heap

    def _map_simulated_losses(self):

        weights_index = np.arange(0,self.kmldata.current_weights.shape[1])
        heap=[]
        for i in weights_index:

            if self.has_args:
                loss = self.loss_function(self.X,self.y,self.kmldata.current_weights[:,i:i+1].copy(),list(self.args))
            else:
                loss = self.loss_function(self.X,self.y,self.kmldata.current_weights[:,i:i+1].copy())
            #make sure the loss is not infinity (this might be a problem for logistic regression)
            if abs(loss)<np.float('inf'):
                heap.append((loss,i))
        return heap


    def _sort_parameters_on_loss(self):
        self.kmldata.current_weights = self.parameter_transform(self.kmldata.current_weights.copy(),self.args)

        if self.use_ipyparallel:
            heap=self._map_simulated_losses_parallel()
        else:
            heap =self._map_simulated_losses()

        sorted_heap = self.heapsort(heap)

        self.kmldata.current_weights = np.array([self.kmldata.current_weights[:,i]
                                for loss,i in reversed(sorted_heap[:self.analyze_n_parameters])]).T
        self.loss = np.array([loss
                                for loss,i in reversed(sorted_heap[:self.analyze_n_parameters])])
        self._update_parameter_history()


    def prior_sampler_uniform_distribution(self,kmldata):
        random_samples = kmldata.prior_random_samples
        number_of_parameterss = kmldata.number_of_parameters
        return np.random.uniform(low=self.low,high=self.high,size=(number_of_parameterss,
                                                                   random_samples))

    def posterior_sampler_uniform_distribution(self,kmldata):

        random_samples = kmldata.prior_random_samples
        variances = np.var(kmldata.update_history[:,:],axis=1).flatten()
        means = kmldata.best_weight_vector.flatten()

        return np.vstack([np.random.uniform(mu-np.sqrt(sigma*12)/2,mu+np.sqrt(sigma*12)/2,(random_samples)) for sigma,mu in zip(variances,means)])



    def _map_losses_parallel(self,w_list):

        weights_index = np.arange(0,w_list.shape[1])
        weights_index_split = np.array_split(weights_index,self.num_cores)

        mydict=dict(weights=w_list)
        self.dview.push(mydict)

        rX = [ipp.Reference('X') for i in range(self.num_cores)]
        ry = [ipp.Reference('y') for i in range(self.num_cores)]
        rweights = [ipp.Reference('weights') for i in range(self.num_cores)]
        rloss_function = [ipp.Reference('loss_function') for i in range(self.num_cores)]
        rargs = [ipp.Reference('args') for i in range(self.num_cores)]
        heap_array = self.dview.map_sync(get_parallel_losses,rloss_function,rX,ry,rweights,rargs,weights_index_split)
        heap =  []
        for val in heap_array:
            heap.extend(val)
        return heap

    def _map_losses(self,w_list):
        weights_index = np.arange(0,w_list.shape[1])
        heap=[]
        for i in weights_index:

            if self.has_args:
                loss = self.loss_function(self.X,self.y,w_list[:,i:i+1],list(self.args))
            else:
                loss = self.loss_function(self.X,self.y,w_list[:,i:i+1])
            #make sure the loss is not infinity (this might be a problem for logistic regression)
            if abs(loss)<np.float('inf'):
                heap.append((loss,i))
        return heap


    def _cycle(self, wh, lh, P=0.1, I=0.1, D=0):


        w = self.kmldata.update_history[:,-1:]

        sw = (self.kmldata.loss_history-np.mean(self.kmldata.loss_history))/np.std(self.kmldata.loss_history)
        sw = (np.exp(-1*sw)/(1+np.exp(-1*sw)))*self.update_volatility
        sw = np.nan_to_num(sw)
        model = linear_model.LinearRegression()

        if np.sum(sw)==0:
            raise CustomException("Not enough variation in sampled parameters, try increasing the number of simulations or volatility")

        model.fit(self.kmldata.update_history.T,self.kmldata.loss_history,sample_weight=sw)
        coefs = model.coef_
        model = None

        effect_size = np.abs(coefs.flatten()*w.flatten())
        best_n_effects  = np.argsort(effect_size)[-self.update_volume:]
        mutation_list = np.concatenate([w for _ in range(best_n_effects.shape[0])],axis=1)[:,:,np.newaxis]
        mutation_list = np.concatenate([mutation_list for _ in range(self.strokes)],axis=2)

        count=0

        for i_ in best_n_effects:

            w_tmp  = w.copy()
            pid = parameter_search(P,I,D)

            pid.SetPoint = self.kmldata.update_history[i_][-1]
            feedback=-self.update_volatility*coefs[i_]+self.kmldata.update_history[i_][-1]

            pid.windup_guard = np.exp(coefs[i_]*self.kmldata.update_history[i_][-1] /np.abs(self.kmldata.update_history[i_][-1]))

            updates = np.zeros(self.strokes)
            for p in range(0,self.strokes):
                updates[p] = feedback
                pid.update(feedback)
                output = pid.output
                feedback += output


            mutation_list[i_,count,:] = updates

            count+=1


        mutation_list = self.parameter_transform(mutation_list,self.args)
        successful_mutations = self._multiple_fire(mutation_list)
        self._breed(successful_mutations)

    def _multiple_fire(self,mutation_list):


        w = self.kmldata.update_history[:,-1].reshape((self.kmldata.update_history.shape[0],1))
        loss_1 = self.kmldata.loss_history[-1]
        for p in range(0,self.strokes):
            misfire=True
            if self.use_ipyparallel and mutation_list.shape[1]>=16:
                heap=self._map_losses_parallel(mutation_list[:,:,p])
            else:
                heap =self._map_losses(mutation_list[:,:,p])

            sorted_heap = reversed(self.heapsort(heap))
            reduced_heap = [(loss_2,i) for loss_2,i in sorted_heap if (loss_2<loss_1) and abs(loss_2)<np.float('inf')]
            n_succesful_mutations = len(reduced_heap)
            successful_mutations = np.zeros((w.shape[0],n_succesful_mutations))
            count = 0
            for loss_2,i_ in reduced_heap:

                w = mutation_list[:,i_,p:p+1]
                self.kmldata.update_history = np.column_stack((self.kmldata.update_history,w))
                self.kmldata.update_history = self.kmldata.update_history[:,1:]

                self.kmldata.loss_history = np.hstack((self.kmldata.loss_history,np.array([loss_2])))
                self.kmldata.loss_history = self.kmldata.loss_history[1:]
                loss_1=loss_2
                misfire=False

                successful_mutations[:,count:count+1] = w
                count+=1
                self.kmldata.increment_number_of_updates()

            if misfire==False:
                break


        return successful_mutations


    def _breed(self,successful_mutations):

        N = successful_mutations.shape[1]
        if N<=1:
            return 0
        loss_1 = self.kmldata.loss_history[-1]
        n_breed_combos = (N+1)*N//2 - N
        breed_list = np.zeros((successful_mutations.shape[0],n_breed_combos))
        count = 0
        for i in range(successful_mutations.shape[1]):
            for j in range(successful_mutations.shape[1]):
                if i<=j:
                    continue
                breed_list[:,count] =0.5*successful_mutations[:,i]+0.5*successful_mutations[:,j]
                count+=1

        breed_list = self.parameter_transform(breed_list,self.args)
        if self.use_ipyparallel and breed_list.shape[1]>=16:
            heap=self._map_losses_parallel(breed_list)
        else:
            heap =self._map_losses(breed_list)

        sorted_heap = reversed(self.heapsort(heap))
        reduced_heap = [(loss_2,i) for loss_2,i in sorted_heap if (loss_2<loss_1) and abs(loss_2)<np.float('inf')]
        for loss_2,i in reduced_heap:

            w = breed_list[:,i:i+1]
            self.kmldata.update_history = np.column_stack((self.kmldata.update_history,w))
            self.kmldata.update_history = self.kmldata.update_history[:,1:]

            self.kmldata.loss_history = np.hstack((self.kmldata.loss_history,np.array([loss_2])))
            self.kmldata.loss_history = self.kmldata.loss_history[1:]
            self.kmldata.increment_number_of_updates()

    def optimize(self,print_feedback=False):

        self.kmldata.current_weights = self.prior_sampler(self.kmldata)
        self.kmldata.reset_histories()
        convergence = np.ones(self.kmldata.number_of_parameters)*3
        loss_per_change = 1e6

        proportional = np.linspace(0.0,0.2,self.update_volume)
        integral = np.linspace(0.0,0.2,self.update_volume)

        for n in range(0,self.number_of_cycles):

            self._sort_parameters_on_loss()

            self._cycle(                           self.kmldata.update_history[:,:],
                                                  self.kmldata.loss_history[:],
                                                  self.P,
                                                  self.I,
                                                  self.D)


            self.loss_by_cycle.append(self.kmldata.loss_history[-1])
            self.update_by_cycle.append(self.kmldata.update_history[:,-1])

            self.kmldata.update_by_cycle = np.array(self.update_by_cycle)
            self.kmldata.loss_by_cycle = np.array(self.loss_by_cycle)
            self.kmldata.best_weight_vector = self.kmldata.update_by_cycle[np.where(self.kmldata.loss_by_cycle==np.min(self.kmldata.loss_by_cycle))[0]]


            if n>0:
                loss_per_change = np.abs((self.kmldata.loss_by_cycle[-1]-self.kmldata.loss_by_cycle[-2])/self.kmldata.loss_by_cycle[-2])

            if loss_per_change<self.min_loss_per_change:
                return True

            #current only accepts parameter vector
            if self.kmldata.best_weight_vector.shape[0]>1:
                self.kmldata.best_weight_vector = self.kmldata.best_weight_vector[0]

            if n>=10:
                convergence = (self.kmldata.best_weight_vector-np.mean(
                                self.kmldata.update_by_cycle[-10:,:],axis=0))/(np.std(
                                self.kmldata.update_by_cycle[-10:,:],axis=0))
                nans = np.isnan(convergence)
                convergence[nans] = 0
                convergence = np.abs(convergence)

            if np.all(np.abs(convergence)<self.convergence_z_score):
                return True


            new_weights = self.random_sampler(self.kmldata)

            if new_weights is not None:
                self.kmldata.current_weights = new_weights

            self.kmldata.increment_cycle_number()
            gc.collect()

            if self.use_ipyparallel:
                self.dview.execute('gc.collect()')

        return False

    def change_random_sampler(self,fcn):
        self.random_sampler = fcn

    def change_prior_sampler(self,fcn):
        self.prior_sampler = fcn

    def change_parameter_transform(self,fcn):
        self.parameter_transform = fcn

    def get_param_by_iter(self):
        return np.array(self.kmldata.update_by_cycle)

    def get_loss_by_iter(self):
        return np.array(self.kmldata.loss_by_cycle)

    def get_best_param(self):
        params = self.get_param_by_iter()
        errors = self.get_loss_by_iter()
        best_w_arr = errors.argsort()[0]
        w = params[best_w_arr]
        return w

    def get_update_history(self):
        return self.kmldata.update_history


class parameter_search:

    def __init__(self, P=0.99, I=0.0, D=0.0):

        self.Kp = P
        self.Ki = I
        self.Kd = D

        self.sample_time = 0.00
        self.current_time = 0.0000
        self.last_time = self.current_time

        self.clear()

    def clear(self):

        self.SetPoint = 0.0

        self.PTerm = 0.0
        self.ITerm = 0.0
        self.DTerm = 0.0
        self.last_error = 0.0

        # Windup Guard
        self.int_error = 0.0
        self.windup_guard = 1

        self.output = 0.0

    def update(self, feedback_value):

        error = self.SetPoint - feedback_value

        self.current_time = self.current_time+0.0005
        delta_time = self.current_time - self.last_time
        delta_error = error - self.last_error

        if (delta_time >= self.sample_time):
            self.PTerm = self.Kp * error
            self.ITerm += error * delta_time

            if (self.ITerm < -self.windup_guard):
                self.ITerm = -self.windup_guard
            elif (self.ITerm > self.windup_guard):
                self.ITerm = self.windup_guard

            self.DTerm = 0.0
            if delta_time > 0:
                self.DTerm = delta_error / delta_time

            # Remember last time and last error for next calculation
            self.last_time = self.current_time
            self.last_error = error

            self.output = self.PTerm + (self.Ki * self.ITerm) + (self.Kd * self.DTerm)

    def setKp(self, proportional_gain):
        self.Kp = proportional_gain

    def setKi(self, integral_gain):
        self.Ki = integral_gain

    def setKd(self, derivative_gain):
        self.Kd = derivative_gain

    def setWindup(self, windup):
        self.windup_guard = windup

    def setSampleTime(self, sample_time):
        self.sample_time = sample_time

class KernelML():

    def __init__(self,
                 prior_sampler_fcn=None,
                 posterior_sampler_fcn=None,
                 intermediate_sampler_fcn=None,
                 mini_batch_sampler_fcn=None,
                 parameter_transform_fcn=None,
                 batch_size=None):

        self.prior_sampler_fcn = prior_sampler_fcn
        self.posterior_sampler_fcn = posterior_sampler_fcn
        self.intermediate_sampler_fcn = self.intermediate_sampler_uniform_distribution
        if intermediate_sampler_fcn is not None:
            self.intermediate_sampler_fcn = intermediate_sampler_fcn
        self.mini_batch_sampler_fcn = self.mini_batch_random_choice
        if mini_batch_sampler_fcn is not None:
            self.mini_batch_sampler_fcn = mini_batch_sampler_fcn
        self.parameter_transform_fcn = parameter_transform_fcn
        self.w = []
        self.batch_size=batch_size
        self.ipyparallel=False
        self.kmldata=None

    class KMLData():

        def __init__(self,prior_random_samples,posterior_random_samples,number_of_parameters,args):
            self.current_weights = np.array([])
            self.best_weight_vector = np.array([])
            self.loss_history = []
            self.update_history = []
            self.update_by_cycle = []
            self.loss_by_cycle = []
            self.loss_by_realization = []
            self.update_by_realization = []
            self.realization_number = 1
            self.cycle_number = 1
            self.number_of_updates = 1
            self.args = args
            self.prior_random_samples = prior_random_samples
            self.posterior_random_samples = posterior_random_samples
            self.number_of_parameters = number_of_parameters

        def increment_realization_number(self):
            self.realization_number = self.realization_number+1

        def increment_cycle_number(self):
            self.cycle_number = self.cycle_number+1

        def increment_number_of_updates(self):
            self.number_of_updates = self.number_of_updates+1

        def reset_histories(self):
            self.loss_history = []
            self.update_history = []
            self.cycle_number=1




    def use_ipyparallel(self,dview):
        self.dview= dview
        self.ipyparallel=True

    def load_kmldata(self,kmldata):
        self.kmldata = kmldata

    def intermediate_sampler_uniform_distribution(self,kmldata):

        random_samples = kmldata.prior_random_samples
        variances = np.var(kmldata.update_history[:,:],axis=1).flatten()
        means = kmldata.update_history[:,-1].flatten()

        return np.vstack([np.random.uniform(mu-np.sqrt(sigma*12)/4,mu+np.sqrt(sigma*12)/4,(random_samples)) for sigma,mu in zip(variances,means)])

    def mini_batch_random_choice(self,X,y,batch_size):
        all_samples = np.arange(0,X.shape[0])
        rand_sample = np.random.choice(all_samples,size=batch_size,replace=False)
        X_batch = X[rand_sample]
        y_batch = y[rand_sample]
        return X_batch,y_batch

    def optimize(self,X,y,loss_function,number_of_parameters,args=[],
            number_of_realizations=1,
            number_of_cycles=20,
            update_volume=10,
            number_of_random_simulations = 1000,
            update_volatility = 1,
            min_loss_per_change=0,
            convergence_z_score=1,
            analyze_n_parameters=None,
            update_magnitude=None,
            prior_random_samples=None,
            posterior_random_samples=None,
            prior_uniform_low=-1,
            prior_uniform_high=1,
            print_feedback=True):

        fresh_start = False
        if self.kmldata is None:
            fresh_start = True
            self.kmldata = self.KMLData(prior_random_samples,posterior_random_samples,number_of_parameters,args)

        parameters = np.zeros((number_of_realizations,number_of_parameters))
        losses = np.zeros((number_of_realizations,1))
        for run in range(number_of_realizations):

            start_time = time.time()
            if self.batch_size is None:
                X_batch = X
                y_batch = y
            else:
                X_batch,y_batch = self.mini_batch_sampler_fcn(X,y,self.batch_size)

            self.model = kernel_optimizer(X_batch,y_batch,loss_function,number_of_parameters=number_of_parameters,
                                      args=args,model=self.kmldata)


            #random simulation parameters
            self.model.default_random_simulation_params(prior_uniform_low=prior_uniform_low
                                                   ,prior_uniform_high=prior_uniform_high
                                                   ,prior_random_samples=prior_random_samples
                                                   ,random_sample_num=posterior_random_samples)


            self.model.adjust_optimizer(number_of_cycles=number_of_cycles,
                                            number_of_random_simulations=number_of_random_simulations,
                                                update_volatility=update_volatility,
                                                analyze_n_parameters=analyze_n_parameters,
                                                update_volume=update_volume)

            if self.ipyparallel:
                self.model.init_ipyparallel(self.dview)
            if self.prior_sampler_fcn is not None:
                self.model.change_prior_sampler(self.prior_sampler_fcn)
            if self.posterior_sampler_fcn is not None:
                self.model.change_random_sampler(self.posterior_sampler_fcn)
            if self.parameter_transform_fcn is not None:
                self.model.change_parameter_transform(self.parameter_transform_fcn)

            self.model.adjust_convergence_z_score(convergence_z_score)
            self.model.adjust_min_loss_per_change(min_loss_per_change)

            if run==0 and fresh_start==False:
                self.model.change_prior_sampler(self.intermediate_sampler_fcn)

            if run>0:
                #this overrides the default simulation function
                #we will use the best parameter set from the previous run as the priors for the current
                self.kmldata = self.model.kmldata
                self.model.change_prior_sampler(self.intermediate_sampler_fcn)

            #fix plot_feedback
            self.model.optimize(print_feedback=print_feedback)

            params = self.model.get_param_by_iter()
            errors = self.model.get_loss_by_iter()
            best_w_arr = errors.argsort()[0]
            loss = errors[best_w_arr]
            w = params[best_w_arr].T

            self.w = w
            parameters[run] = w
            losses[run] = loss
            self.kmldata.increment_realization_number()
            end_time = time.time()

            if print_feedback:
                print('realization',run,'loss',loss,'time',end_time-start_time)

        self.kmldata.loss_by_realization = losses
        self.kmldata.update_by_realization = parameters
        return self



from setuptools import setup, Extension, find_packages
from distutils.core import setup
from Cython.Build import cythonize


from codecs import open
import os
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'kernelml-README.md'), encoding='utf-8') as f:
    long_description = f.read()


# setup(
#     name = "Kernel Machine Learning",
#     ext_modules = cythonize('kernelml.pyx'),  # accepts a glob pattern
# )


def find_pyx(path='.'):
    pyx_files = []
    for root, dirs, filenames in os.walk(path):
        for fname in filenames:
            if fname.endswith('.pyx'):
                pyx_files.append(os.path.join(root, fname))
                
    return pyx_files

def find_c(path='.'):
    c_files = []
    for root, dirs, filenames in os.walk(path):
        for fname in filenames:
            if fname.endswith('.c'):
                c_files.append(os.path.join(root, fname))
    return c_files

cythonize(find_pyx('hdre'))

setup(
      # Information
      name = "hdre",
      version = "1.01",
      description='generalized machine learning algorithm for complex loss functions and non-linear coefficients',
      url = "https://github.com/Freedomtowin/kernelml",
      author = "Rohan Kotwani",
      email = "rohankotwani@gmail.com",
      
      
      license = "MIT",
      classifiers=[
                   "Development Status :: 4 - Beta",
                   
                   # Indicate who your project is intended for
                   "Intended Audience :: Developers",
                   "Intended Audience :: Science/Research",
                   "Topic :: Software Development",
                   "Topic :: Scientific/Engineering",
                   
                   # Pick your license as you wish
                   'License :: OSI Approved :: MIT License',
                   
                   # Specify the Python versions you support here. In particular, ensure
                   # that you indicate whether you support Python 2, Python 3 or both.
                   'Programming Language :: Python :: 3'
                   ],
      keywords = "high density region estimator - kernel machine learning nonlinear",
      #    packages=find_packages(exclude=['contrib', 'docs', 'tests']),  # Required
      packages=['hdre', 'hdre.hdre_bycython','hdre.hdre_bycython.hdr_helpers_bycython'],  # Required
      install_requires = ["numpy","cython","numba"],
      python_requires = ">=3.6",
      
      #    # Build instructions
      
      ext_modules = [
                     Extension("hdre.hdre_bycython.hdr_helpers_bycython.hdr_helpers",
                               find_c('hdre/hdre_bycython/hdr_helpers_bycython'),
                               include_dirs=[])
                               ]
                               
      )
